package br.ime.usp.mac5855.otimizacao;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Classe que implementa algoritmos de otimizacao.
 * 
 * @see {@link br.ime.usp.mac5855.otimicazao.FuncaoDeCusto}
 * */
public class Otimizador {

	private Random random = new Random(System.currentTimeMillis());

	/**
	 * Executa otimizacao randomica utilizando a funcao de custo <code>funcao</code>,
	 * para o dominio <code>dominio</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @param limite A quantidade de tentativas a serem feitas.
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoRandomica(Dominio dominio, FuncaoDeCusto funcao, int limite){	
		int [] solucao = null;
		double melhor = Double.MAX_VALUE;
		
		for(int i = 0; i < limite; i++){
			int [] tentativa = geraTentativa(dominio);
			double custo = funcao.calculaCusto(tentativa);
			if(custo < melhor){
				melhor  = custo;
				solucao = tentativa;
			}			
		}				
		return new Resultado(solucao, melhor);
	}
	
	/**
	 * Executa otimizacao usando a estrategia gulosa Hill Climbing, fazendo uso
	 * da funcao de custo <code>funcao</code>, para o dominio <code>dominio</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoGulosa(Dominio dominio, FuncaoDeCusto funcao){
		int [] solucao = geraTentativa(dominio);
		double melhor = 0.0;
		boolean melhorado = false;
		while(true){
			List<int[]> vizinhos = geraVizinhos(dominio, solucao);
			melhor = funcao.calculaCusto(solucao);
			melhorado = false;
			double custo = melhor;
			for(int i = 0; i < vizinhos.size(); i++){
				custo = funcao.calculaCusto(vizinhos.get(i));
				if(custo < melhor){
					melhor  = custo;
					solucao = vizinhos.get(i);
					melhorado = true;
				}
			}
			if(!melhorado){
				break;
			}
		}
		return new Resultado(solucao, melhor);
	}

	/**
	 * Executa otimizacao usando a estrategia de aquecer e esfriar (simulated annealing) fazendo uso
	 * da funcao de custo <code>funcao</code>, para o dominio <code>dominio</code>, iniciando na temperatura
	 * <code>temperatura</code> e coeficiente de resfriamento <code>esfriamento</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @param temperatura A temperatura inicial (positiva)
	 * @param esfriamento O coeficiente de esfriamento (entre ]0,1[)
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoPorSimulacaoDeAquecimentoEsfriamento(Dominio dominio, FuncaoDeCusto funcao, double temperatura, double esfriamento){
		if(temperatura <= 0.0 || esfriamento >= 1.0 || esfriamento <= 0.0){
			throw new IllegalArgumentException("Temperatura deve ser positiva e esfriamento deve estar no intervalo ]0.0,1.0[");
		}
		int [] solucao = geraTentativa(dominio);
		double melhor = Double.MAX_VALUE;
		while(temperatura > 0.1){
			int indice  = getRandomico(0, dominio.getTamanho());
			int direcao = (indice % 2 == 0) ? -1 : 1;
			int [] tentativa = solucao.clone();
			tentativa[indice] += direcao;
			
			if(tentativa[indice] < dominio.getLimiteInferior(indice)){
				tentativa[indice] = dominio.getLimiteInferior(indice);
			}
			else if(tentativa[indice] > dominio.getLimiteSuperior(indice)){
				tentativa[indice] = dominio.getLimiteSuperior(indice);
			}
			melhor = funcao.calculaCusto(solucao);
			double custo = funcao.calculaCusto(tentativa);
			double  probabilidade = Math.pow(Math.E, (-custo - melhor)/temperatura);
			if(custo < melhor || probabilidade > random.nextDouble()){
				solucao = tentativa;
				melhor = custo;
			}
			temperatura*=esfriamento;
			
		}		
		return new Resultado(solucao, melhor);
	}
	
	/**
	 * Executa otimizacao usando a estrategia de aquecer e esfriar (simulated annealing) fazendo uso
	 * da funcao de custo <code>funcao</code>, para o dominio <code>dominio</code>, iniciando na temperatura
	 * <code>1000.0</code> e coeficiente de resfriamento <code>0.95</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoPorSimulacaoDeAquecimentoEsfriamento(Dominio dominio, FuncaoDeCusto funcao){
		return otimizacaoPorSimulacaoDeAquecimentoEsfriamento(dominio, funcao, 1000.0, .95);
	}
	
	private List<int[]> geraVizinhos(Dominio dominio, int[] solucao) {
		List<int[]> vizinhos = new ArrayList<int[]>();
		for(int i = 0; i < dominio.getTamanho(); i++){
			if(solucao[i] > dominio.getLimiteInferior(i)){
				int [] tentativa = new int[dominio.getTamanho()];
				for(int k = 0; k < i; k++){
					tentativa[k] = solucao[k];
				}
				tentativa[i] = solucao[i] - 1;
				for(int k = i+1; k < dominio.getTamanho(); k++){
					tentativa[k] = solucao[k];
				}
				vizinhos.add(tentativa);
			}
			if(solucao[i] < dominio.getLimiteSuperior(i)){
				int [] tentativa = new int[dominio.getTamanho()];
				for(int k = 0; k < i; k++){
					tentativa[k] = solucao[k];
				}
				tentativa[i] = solucao[i] + 1;
				for(int k = i+1; k < dominio.getTamanho(); k++){
					tentativa[k] = solucao[k];
				}
				vizinhos.add(tentativa);
			}								
		}
		return vizinhos;
	}
	
	private int[] geraTentativa(Dominio dominio) {
		int [] resultado = new int[dominio.getTamanho()];
		for(int i = 0; i < dominio.getTamanho(); i++){
			resultado[i] = getRandomico(dominio.getLimiteInferior(i),
										dominio.getLimiteSuperior(i) + 1);
		}
		return resultado;
	}

	private int getRandomico(int limiteInferior, int limiteSuperior) {
		return (random.nextInt(limiteSuperior) % limiteSuperior) + limiteInferior;
	}

	/**
	 * Executa otimizacao randomica utilizando a funcao de custo <code>funcao</code>,
	 * para o dominio <code>dominio</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoRandomica(Dominio dominio, FuncaoDeCusto funcao){
		return otimizacaoRandomica(dominio, funcao, 1000);
	}

	/**
	 * Executa otimizacao atraves de uma estrategia de mutacoes e cruzamentos entre seres da populacao
	 * gerada. O custo eh calculado pela funcao de custo <code>funcao</code>,
	 * para o dominio <code>dominio</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoGenetica(Dominio dominio, FuncaoDeCusto funcao){
		return otimizacaoGenetica(dominio, funcao, 50, 0.2, 0.2, 100);
	}

	/**
	 * Executa otimizacao atraves de uma estrategia de mutacoes e cruzamentos entre seres da populacao
	 * gerada. O custo eh calculado pela funcao de custo <code>funcao</code>,
	 * para o dominio <code>dominio</code>.
	 * 
	 * @param dominio O dominio dos valores a ser utilizado.
	 * @param funcao A funcao de custo a ser usada.
	 * @param tamanho O tamanho inicial da populacao.
	 * @param propabilidadeDeMutacao A probabilidade de um novo membro da populacao ser uma mutacao, ao inves de ser resultado um cruzamento.
	 * @param elite Fracao da populacao que sera considerada uma boa solucao e seguira para as proximas geracoes.
	 * @param qtdeDeGeracoes O numero de geracoes que serao criadas.
	 * @return o melhor resultado do dominio calculado na funcao de custo.
	 * */
	public Resultado otimizacaoGenetica(Dominio dominio, FuncaoDeCusto funcao, 
			int tamanhoDaPopulacao, double probabilidadeDeMutacao, double elite, int qtdeDeGeracoes){
		if(tamanhoDaPopulacao <= 0 || probabilidadeDeMutacao <= 0.0 || elite <= 0.0 || qtdeDeGeracoes < 0){
			throw new IllegalArgumentException("Os valores devem ser estritamente positivos.");
		}
		ArrayList<int[]> populacao = geraPopulacaoInicial(dominio, tamanhoDaPopulacao);		
		int eliteQtde = (int)(elite * tamanhoDaPopulacao);
		for(int i = 0; i < qtdeDeGeracoes; i++){
			List<Par> lista = new ArrayList<Par>();
			for(int j = 0; j < populacao.size(); j++){
				int [] instancia = populacao.get(j);			
				lista.add(new Par(funcao.calculaCusto(instancia), instancia));
			}
			populacao = getMelhoresDaPopulacao(lista, eliteQtde);			
			while(populacao.size() < tamanhoDaPopulacao){
				if(random.nextDouble() < probabilidadeDeMutacao){
					int indice = getRandomico(0, eliteQtde);
					populacao.add(executaMutacao(populacao.get(indice), dominio));
				}else{
					int indice1 = getRandomico(0, eliteQtde);
					int indice2 = getRandomico(0, eliteQtde);
					populacao.add(executaCruzamento(populacao.get(indice1), populacao.get(indice2)));
				}
			}
		}		
		return new Resultado(populacao.get(0), funcao.calculaCusto(populacao.get(0)));
	}

	private class Par implements Comparable<Par>{
		double valor;
		
		int [] info;
		
		Par(double valor, int [] info){
			this.valor = valor;
			this.info = info;
		}
		
		public int compareTo(Par outro) {
			return (int)(this.valor - outro.valor);
		}
		
	}
	
	private ArrayList<int[]> getMelhoresDaPopulacao(List<Par> lista, int eliteQtde) {		
		ArrayList<int[]> resultado = new ArrayList<int[]>();
		for(int i = 0; i < eliteQtde; i++){
			resultado.add(lista.get(i).info);			
		}		
		return resultado;
	}

	private int[] executaMutacao(int [] organismo, Dominio dominio){
		int posicao = getRandomico(0, dominio.getTamanho());
		if(random.nextDouble() < 0.5 && organismo[posicao] > dominio.getLimiteInferior(posicao)){
			int [] resultado = new int[dominio.getTamanho()];
			for(int i = 0; i < posicao; i++){
				resultado[i] = organismo[i];
			}
			resultado[posicao] = organismo[posicao] - 1;
			for(int i = posicao + 1; i < organismo.length; i++){
				resultado[i] = organismo[i];
			}
			return resultado;
		}else if(organismo[posicao] < dominio.getLimiteSuperior(posicao)){
			int [] resultado = new int[dominio.getTamanho()];
			for(int i = 0; i < posicao; i++){
				resultado[i] = organismo[i];
			}
			resultado[posicao] = organismo[posicao] + 1;
			for(int i = posicao + 1; i < organismo.length; i++){
				resultado[i] = organismo[i];
			}
			return resultado;
		}		
		return organismo;
	}
	
	private int[] executaCruzamento(int [] organismo1, int [] organismo2){
		int posicao = getRandomico(1, organismo1.length - 1);
		int [] resultado = new int[organismo1.length];
		for(int i = 0; i < posicao; i++){
			resultado[i] = organismo1[i];
		}
		for(int i = posicao; i < resultado.length; i++){
			resultado[i] = organismo2[i];
		}
		return resultado;
	}
	
	private ArrayList<int[]> geraPopulacaoInicial(Dominio dominio, int tamanhoDaPopulacao) {
		ArrayList<int[]> lista = new ArrayList<int[]>();
		for(int i = 0; i < tamanhoDaPopulacao; i++){
			lista.add(geraTentativa(dominio));
		}		
		return lista;
	}
}
